/* This file is generated by glib-genmarshal, do not modify it. This code is licensed under the same license as the containing project. Note that it links to GLib, so must comply with the LGPL linking clauses. */
#ifndef __GS_THEME_ENGINE_MARSHAL_MARSHAL_H__
#define __GS_THEME_ENGINE_MARSHAL_MARSHAL_H__

#include <glib-object.h>

G_BEGIN_DECLS

/* BOOLEAN:BOXED (gs-theme-engine-marshal.list:1) */
extern
void gs_theme_engine_marshal_BOOLEAN__BOXED (GClosure     *closure,
                                             GValue       *return_value,
                                             guint         n_param_values,
                                             const GValue *param_values,
                                             gpointer      invocation_hint,
                                             gpointer      marshal_data);


G_END_DECLS

#endif /* __GS_THEME_ENGINE_MARSHAL_MARSHAL_H__ */
